import os
import os.path
import sys
import pwd
import re


def become_user(user):
    if os.geteuid() != pwd.getpwnam(user).pw_uid:
        # flush output buffers. Otherwise the output before the
        # become_root()-call might be never shown to the user
        sys.stdout.flush()
        sys.stderr.flush()

        # resolve symlinks, so the path given in sudo-config matches
        realme = os.path.realpath(sys.argv[0])

        # -n disables password prompt, when sudo isn't configured properly
        os.execv('/usr/bin/sudo', ['/usr/bin/sudo', '-u', user, '-n', '--', realme] + sys.argv[1:])
    else:
        sudouser = os.environ.get('SUDO_USER', None)
        # $SUDO_USER is set directly by sudo, so users should not be alble
        # to trick here. Better be safe, than sorry, though.
        if sudouser and re.match('^[a-z][a-z0-9]{0,20}$', sudouser):
            return sudouser
        else:
            raise ValueError('invalid username: "{}"'.format(sudouser))


def check_root():
    return os.geteuid() == 0



from __future__ import print_function

import argparse
import sys
import os.path
import inspect

import six

from .runners.ansiblerunner import AnsibleRunner
from .root import become_user, check_root


class Paternoster:
    def __init__(self, runner_parameters, parameters, success_msg='executed successfully', runner_class=AnsibleRunner):
        self._parameters = parameters
        self._success_msg = success_msg
        self._sudo_user = None
        self._runner = runner_class(**runner_parameters)

    def _find_param(self, fname):
        """ look for a parameter by either its short- or long-name """
        for name, short, param in self._parameters:
            if name == fname or short == fname:
                return (name, short, param)

        raise KeyError('Parameter {0} could not be found'.format(fname))

    def _check_type(self, argParams):
        """ assert that an argument does not use a string type opposed to an restricted_str, else raise a ValueError """
        action_whitelist = ('store_true', 'store_false', 'store_const', 'append_const', 'count')
        action = argParams.get('action', 'store')

        if 'type' not in argParams and action not in action_whitelist:
            raise ValueError('a type must be specified for each user-supplied argument')

        type = argParams.get('type', str)
        is_str_type = inspect.isclass(type) and issubclass(type, six.string_types)

        if is_str_type and action not in action_whitelist:
            raise ValueError('restricted_str instead of str or unicode must be used for all string arguments')

    def _build_argparser(self):
        parser = argparse.ArgumentParser(add_help=False)
        requiredArgs = parser.add_argument_group('required arguments')
        optionalArgs = parser.add_argument_group('optional arguments')

        optionalArgs.add_argument(
            '-h', '--help', action='help', default=argparse.SUPPRESS,
            help='show this help message and exit'
        )

        for name, short, param in self._parameters:
            argParams = param.copy()
            argParams.pop('depends', None)
            argParams.pop('positional', None)

            self._check_type(argParams)

            if param.get('positional', False):
                paramName = [name]
            else:
                paramName = ['-' + short, '--' + name]

            if param.get('required', False) or param.get('positional', False):
                requiredArgs.add_argument(*paramName, **argParams)
            else:
                optionalArgs.add_argument(*paramName, **argParams)

        optionalArgs.add_argument(
            '-v', '--verbose', action='count', default=0,
            help='run with a lot of debugging output'
        )

        return parser

    def _check_arg_dependencies(self, parser, args):
        for name, short, param in self._parameters:
            if 'depends' in param and getattr(args, name, None) and not getattr(args, param['depends'], None):
                parser.error(
                    'argument --{} (-{}) requires --{} (-{}) to be present.'.format(
                        name, short, param['depends'], self._find_param(param['depends'])[1]
                    )
                )

    def check_root(self):
        if not check_root():
            print('This script can only be used by the root user.', file=sys.stderr)
            sys.exit(1)

    def become_user(self, user):
        try:
            self._sudo_user = become_user(user)
        except ValueError as e:
            print(e, file=sys.stderr)
            sys.exit(1)

    def auto(self, become_root=False, become_user=None, check_root=False):
        if sum([bool(become_root), bool(become_user), bool(check_root)]) > 1:
            raise ValueError('check_root, become_user and become_root cannot be supplied together')

        if check_root:
            self.check_root()
        if become_user:
            self.become_user(become_user)
        if become_root:
            self.become_user('root')
        self.parse_args()
        status = self.execute()
        sys.exit(0 if status else 1)

    def parse_args(self, args=None):
        parser = self._build_argparser()

        args = parser.parse_args(args)
        self._check_arg_dependencies(parser, args)

        self._parsed_args = args

    def _get_runner_variables(self):
        if self._sudo_user:
            yield ('sudo_user', self._sudo_user)

        yield ('script_name', os.path.basename(sys.argv[0]))

        for name in vars(self._parsed_args):
            yield ('param_' + name, getattr(self._parsed_args, name))

    def execute(self):
        status = self._runner.run(self._get_runner_variables(), self._parsed_args.verbose)
        if status and self._success_msg:
            print(self._success_msg)
        return status



from __future__ import print_function

import sys
import os
import os.path
from collections import namedtuple

# Ansible loads the ansible.cfg in the following order. Each ansible.cfg
# automatically overwrites all values of the former ones.
#
#  1. Path given in $ANSIBLE_CONFIG
#  2. ansible.cfg file in the current directory
#  3. .ansible.cfg file in the home directory
#  4. /etc/ansible/ansible.cfg
#
# Since the current directory is controlled by the user and we don't
# want them to be able to load their own config and thus their own
# ansible modules, we need to counter them by setting the env-variable.
os.environ['ANSIBLE_CONFIG'] = '/etc/ansible/ansible.cfg'

# by default ansible uses "$HOME/.ansible/tmp" as the directory to drop
# its module files. For some reason $HOME is not resolved when using the
# python API directly resuling in a new directory called '$HOME' within
# the paternoster source. This forces the modules to be dropped in /tmp.
os.environ['ANSIBLE_REMOTE_TEMP'] = '/tmp'
os.environ['ANSIBLE_LOCAL_TEMP'] = '/tmp'

# Verbosity within ansbible is controlled by the Display-class. Each and
# every ansible-file creates their own instance of this class, like this:
#
#  try:
#     from __main__ import display
#   except ImportError:
#     from ansible.utils.display import Display
#     display = Display()
#
# This means that the verbosity-parameter of display _always_ default to
# zero. There is no sane way to overwrite this. Within a normal ansible
# setup __main__ corresponds to the current executable (e.g. "ansible-playbook"),
# which creates a Display instance based on the cli parameters (-v, -vv, ...).
#
# This has to happen before anything from ansible is imported!
import __main__
from ansible.utils.display import Display

__main__.display = Display()

import ansible.constants
from ansible.executor.playbook_executor import PlaybookExecutor
from ansible.inventory import Inventory
from ansible.parsing.dataloader import DataLoader
from ansible.plugins.callback import CallbackBase
from ansible.vars import VariableManager


class MinimalAnsibleCallback(CallbackBase):
    """ filters out all ansible messages except for playbook fails and debug-module-calls. """

    def v2_runner_on_failed(self, result, ignore_errors=False):
        if not ignore_errors:
            print(result._result['msg'], file=sys.stderr)

    def v2_runner_item_on_ok(self, result):
        self.v2_runner_on_ok(result)

    def v2_runner_on_ok(self, result):
        result = result._result
        if 'invocation' in result:
            if result['invocation'].get('module_name', None) == 'debug':
                args = result['invocation'].get('module_args', None)
                if 'var' in args:
                    print(result[args['var']])
                if 'msg' in args:
                    print(args['msg'])


class AnsibleRunner:
    def __init__(self, playbook):
        self._playbook = playbook

    def _get_playbook_executor(self, variables, verbose):
        Options = namedtuple('Options',
                             ['connection', 'module_path', 'forks', 'become', 'become_method', 'become_user', 'check',
                              'listhosts', 'listtasks', 'listtags', 'syntax'])

        # -v given to us enables ansibles non-debug output.
        # So -vv should become ansibles -v.
        __main__.display.verbosity = max(0, verbose - 1)

        variable_manager = VariableManager()
        loader = DataLoader()
        inventory = Inventory(loader=loader, variable_manager=variable_manager, host_list=['localhost'])
        variable_manager.set_inventory(inventory)
        # force ansible to use the current python executable. Otherwise
        # it can end up choosing a python3 one (named python) or a different
        # python 2 version
        variable_manager.set_host_variable(inventory.localhost, 'ansible_python_interpreter', sys.executable)

        for name, value in variables:
            variable_manager.set_host_variable(inventory.localhost, name, value)

        pexec = PlaybookExecutor(
            playbooks=[self._playbook],
            inventory=inventory,
            variable_manager=variable_manager,
            loader=loader,
            options=Options(
                connection='local',
                module_path=None,
                forks=1,
                listhosts=False, listtasks=False, listtags=False, syntax=False,
                become=None, become_method=None, become_user=None, check=False
            ),
            passwords={},
        )

        ansible.constants.RETRY_FILES_ENABLED = False

        if not verbose:
            # ansible doesn't provide a proper API to overwrite this,
            # if you're using PlaybookExecutor instead of initializing
            # the TaskQueueManager (_tqm) yourself, like in the offical
            # example.
            pexec._tqm._stdout_callback = MinimalAnsibleCallback()

        return pexec

    def _check_playbook(self):
        if not self._playbook:
            raise ValueError('no playbook given')
        if not os.path.isabs(self._playbook):
            raise ValueError('path to playbook must be absolute')
        if not os.path.isfile(self._playbook):
            raise ValueError('playbook must exist')

    def run(self, variables, verbose):
        self._check_playbook()
        os.chdir(os.path.dirname(self._playbook))
        status = self._get_playbook_executor(variables, verbose).run()
        return True if status == 0 else False



from __future__ import print_function
import sys

try:
    from setuptools import setup
except ImportError:
    print("Paternoster needs setuptools.", file=sys.stderr)
    print("Please install it using your package-manager or pip.", file=sys.stderr)
    sys.exit(1)

setup(name='paternoster',
      version='1.1.1',
      description='Paternoster provides users with the ability to run certain tasks as '
                  'root or another user, while ensuring safety by providing a common '
                  'interface and battle tested parameter parsing/checking.',
      author='uberspace.de',
      author_email='hallo@uberspace.de',
      url='https://github.com/uberspace/paternoster',
      packages=[
          'paternoster',
          'paternoster.runners',
      ],
      install_requires=[
          'tldextract>=2.0.1',
          'ansible==2.2.0.0',
          'six',
      ],
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Intended Audience :: Information Technology',
          'Intended Audience :: System Administrators',
          'Topic :: System :: Systems Administration',
          'Topic :: Security',
          'Topic :: Utilities',
          'Natural Language :: English',
          'Operating System :: POSIX :: Linux',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 2 :: Only',
      ],
      zip_safe=True,
      )
